/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * This module is responsible to lookup queries.
 * @file common/lookup.c
 * @author Christian Grothoff
 **/
#include <stdio.h>
#include <dirent.h>
#include "config.h"
#include "lookup.h"
#include "util/contentdatabase.h"
#include "util/xmalloc.h"

/* ********************* GLOBALS ***************** */

/* handle of the database-file, never closed yet */
static CollisionLevel * rootLevel;


/**
 * names of indexed files
 **/
char ** indexed_files = NULL;

/**
 * Size of the indexed_files list.
 **/
unsigned short indexed_files_size;

/**
 * number of files that are indexed
 **/
static unsigned short indexed_files_count;

/* number of files that are in data/content */
static unsigned int LOOKUP_filesStored_;

/* what is the priority of the least important content
   that we are storing */
static int LOOKUP_leastImportance_;

/* the hashcode of the least important entry */
static HashCode160 LOOKUP_lip_;

/* the least important entries */
static int LOOKUP_lips_[LOOKUP_LIPS_COUNT];
/* the number of valid entries in lookup-lips */
static int LOOKUP_lips_count_;
/* the current partition of the database for lips */
static int LOOKUP_lips_db_index_;

/* value of the current base value for
   fresh content (used to time-out
   old content) */
static int LOOKUP_age;

/* ************* internal Methods **************** */

/**
 * Cron-job that makes sure the lookup-module is fine
 * by synchronizing the in-memory database with the drive.
 **/
static void cronSyncLookup(CollisionLevel * level);

/**
 * Cron-job that decreases the importance-level of all
 * files by 1. Runs 'not very often'.
 **/
static void cronReduceImportance(void * unused);

/* scan the list of on-demand shared files to
   initialize LOOUP_fileNames_ */
static int scanDatabaseList();

/**
 * Write the given entry to the database. 
 * @return OK on success, SYSERR on error
 **/
static int addEntry_(CollisionLevel * level,
		     ContentIndex * entry);

/**
 * Given an entry in the cache, obtain the matching
 * message(s). Send them to the given host.
 * @param entry the entry to look for
 * @param method a method pointer used for callback; this is the
 *               method to be called if the lookup succeeds. 
 *               must take three arguments: NULL, a pointer
 *               to the CONTENT_Block, and the query.
 **/ 
static void resolveEntry(ContentIndex * entry,
			 int (* method)(HostIdentity*,CONTENT_Block*,
					 HashCode160*));

/**
 * Lookup the given Hash in the Hashtable-cache structure
 * (load blocks as needed).
 * @returns NO if not found, YES if found
 **/
static int findEntry_(CollisionLevel ** level,
		     HashCode160 * query,
		     ContentIndex * entry);

/**
 * Write the given entry to the database.
 **/
static void writeEntry(CollisionLevel * level,
		       ContentIndex * entry,
		       int isNew);

/**
 * Delete an entry and the corresponding file on the
 * drive.
 **/
static void deleteEntry_(CollisionLevel * level,
			 HashCode160 * entry);

/**
 * Find the last important entry and
 * set the LOOKUP_lip_ and LOOKUP_leastImportance_
 * variables accordingly.
 **/
static void findLeastImportant(CollisionLevel * level);

/**
 * Check if the file the query is for is on the
 * local drive. If yes, send the reply and return YES,
 * otherwise return NO.
 * @param query the filename to look for
 * @param method a method pointer used for callback; this is the
 *               method to be called if the lookup succeeds. 
 *               must take three arguments: NULL, a pointer
 *               to the CONTENT_Block, and the query.
 * @return YES if the reply succeeded, else NO
 **/
static int localLookup(HashCode160 * query,		       
		       int (* method)(HostIdentity*,CONTENT_Block*,
				       HashCode160*));


/* ******************* CODE... ****************** */

/**
 * Initialize the lookup-module.
 **/
void initLookup() {  
  CollisionLevel * tmp;
  FileName fileName;
  FileName fileNameDotted;
  int handle;
  int i;
  
  fileName = getAgeFile();
  handle = open(fileName,O_CREAT,S_IREAD|S_IWRITE);
  LOOKUP_age = 0;
  read(handle, 
       &LOOKUP_age,
       sizeof(int));
  close(handle);
  LOOKUP_lips_db_index_ = rand() % LOOKUP_LIPSPART_COUNT;
  LOOKUP_lips_count_ = 0;

  DEBUG("Indexing data/content database... ");
  LOOKUP_leastImportance_ = 0;
  /* scan lookup-file index */
  scanDatabaseList();

  fileName = getDatabaseFile();
  rootLevel = NULL;
  for (i=LOOKUP_LEVELS-1;i>=0;i--) {
    tmp = xmalloc(sizeof(CollisionLevel),
		  "initLookup: collisionLevel");
    tmp->next = rootLevel; 
    tmp->mask = (CACHE_HASH_TABLE_SIZE >> (i*LOOKUP_POWER_FACTOR))-1;
    tmp->offset = i;
    create_mutex(&tmp->lock);
    rootLevel = tmp;
    fileNameDotted = xmalloc(strlen(fileName)+3,
			     "initLookup: database.X filename");
    strcpy(fileNameDotted, 
	   fileName);
    strcat(fileNameDotted, ".0");
    fileNameDotted[strlen(fileNameDotted)-1]+=i;
    tmp->handle = open(fileNameDotted, 
		       O_CREAT|O_RDWR, 
		       S_IRUSR|S_IWUSR);  
    if (tmp->handle < 0)
      errexit("Could not open/create database %s!\n",
	      fileNameDotted);
    xfree(fileNameDotted,
	  "initLookup");
  }
  LOOKUP_filesStored_ = countContentEntries();
  DEBUG("Entries: %d\n", LOOKUP_filesStored_);
}

void fullInitLookup() {
  CollisionLevel * tmp;
  int i;

  if (YES == activeMigration())
    findLeastImportant(rootLevel);
  /* schedule cron job */
  tmp = rootLevel;
  i=5;
  while (tmp != NULL) {
    addCronJob((void (*)(void*))&cronSyncLookup,
	       1000*(i*60+i), 1000*(i*60), 
	       tmp);
    tmp = tmp->next;
    i+=3;
  }
  /* schedule cron job */
  addCronJob(&cronReduceImportance,
	     1000*(6*60*60), 1000*(12*60*60),
	     NULL); /* decrease every 12h, but first time after 6h */
}

/**
 * Shutdown the lookup module.
 **/
void doneLookup() {
  CollisionLevel * tmp;
  int i;

  tmp = rootLevel;
  i=5;
  while (tmp != NULL) {
    delCronJob((void (*)(void*))&cronSyncLookup,
	       1000*(i*60), 
	       tmp);
    tmp = tmp->next;
    i+=3;
  }
  delCronJob(&cronReduceImportance,
	     1000*(12*60*60),
	     NULL); 
  tmp = rootLevel;
  while (rootLevel != NULL) {
    tmp = rootLevel;
    rootLevel = rootLevel->next;
    destroy_mutex(&tmp->lock);
    close(tmp->handle);
    xfree(tmp,
	  "doneLookup, database.X colision level");
  }

}

/**
 * Obtain the number of shared files.
 **/
unsigned int getSharedFileCount() {
  return indexed_files_size;
}

/**
 * Obtain the offset in the file for the given
 * triple-Hash.
 **/
static size_t getOffset(CollisionLevel * level,
			HashCode160 * tripleHash) {
#if VERY_VERBOSE_LOOKUP
  HexName hex;
  hash2hex(tripleHash,
	   &hex);
  printf("computing index at level %d mask %x for hash %s\n",
	 level->offset, level->mask, 
	 &hex);
#endif
  return (((int*)tripleHash)[level->offset] & level->mask)
    * sizeof(ContentIndex);
}

/**
 * Invoke a method on each of the filenames of
 * the indexed files. If the method returns
 * SYSERR, remove the file from the list of
 * indexed files!
 * @param data the last argument to method
 * @return the number of shared files (after changes
 *         caused by this call)
 **/
int forEachIndexedFile(int (*method)(FileName,int)) {
  int i;
  int changed = NO;
  char * fn;

  for (i=0;i<indexed_files_size;i++)
    if (indexed_files[i] != NULL) {
      fn = strdup(indexed_files[i]);
      if (SYSERR == method(fn,
                           i+1)) {
        freeNonNull(indexed_files[i],
                    "file has disappeared");
        indexed_files[i] = NULL;
        changed = YES;
      } 
      xfree(fn, "strduped filename");
    }
  if (YES == changed) {
    /* write changed list to the drive */
    char * fil;
    FILE * handle;

    fil = getSharedFileList();
    handle = fopen(fil, "w+");
    if (handle == NULL) {
      return SYSERR;
    }
    for (i=0;i<indexed_files_size;i++)
      if (indexed_files[i] != NULL)
        fprintf(handle,
                "%s\n",
                indexed_files[i]);
      else
        fprintf(handle,
                "\n");
    fclose(handle);
  }
  return indexed_files_count;
}

/**
 * Check if the file the query is for is on the
 * local drive. If yes, send the reply and return YES,
 * otherwise return NO.
 * @param query the filename to look for (double hash!)
 * @param method a method pointer used for callback; this is the
 *               method to be called if the lookup succeeds. 
 *               must take three arguments: NULL, a pointer
 *               to the CONTENT_Block, and the query.
 * @return YES if success, else NO              
 **/
static int localLookup(HashCode160 * query,
		       int (* method)(HostIdentity*,CONTENT_Block*,
				       HashCode160*)) {
  CONTENT_Block *buf;
  int i,j,k,maxSend;
  unsigned int *randomArray;
  HexName hex;
  hash2hex(query, &hex);

  buf = (CONTENT_Block*)xmalloc(sizeof(CONTENT_Block)*MAX_FSIZE,
      "localLookup: buf");

  DEBUG("QUERY: local lookup for %s.\n",
	&hex);
  j = readContent(query, 
		  MAX_FSIZE * CONTENT_SIZE, 
		  buf);

  if (j <= 0) {
    /* this should not happen, but we're tolerant! */
    INFO("QUERY: local lookup: %s not found!\n",
	  (char*) &hex);
    xfree(buf, "localLookup: buf");
    return NO;
  }
  /* data found on local node! */
  if ((j % CONTENT_SIZE) != 0) {
    hash2hex(query, &hex);
    WARNING("Bad content file: %s\n",
	  (char*)&hex);
    xfree(buf, "localLookup: buf");
    return NO;
  }
  INFO("QUERY: local lookup succeded. Sending reply (%s)!\n",
	(char*)&hex);
  /* send multiple replies if applicable */
  j=j/CONTENT_SIZE;

  /* send as much as we can afford, atleast 1 */
  maxSend=getMaxNetLoad()/(networkUsage()+1);
  if(maxSend<1)
    maxSend=1;
  
  /* select randomly upto maxSend blocks from possible choices */
  randomArray=xmalloc(j*sizeof(unsigned int), 
  				"localLookup: randomArray");

  for (i=0;i<j;i++) {
    randomArray[i]=i;
  }
 
  for (i=0;i<j;i++) {
    int tmp=randomArray[i];
	int tmpIndex=randomi(j-i)+i;

	randomArray[i]=randomArray[tmpIndex];
	randomArray[tmpIndex]=tmp;
  }
  
  for(k=0 ; k<j && k<maxSend ; k++) {
    /*  method = callback; */
    method(NULL, /* comes from local */
	   &buf[randomArray[k]], 
	   query);  
  }

  xfree(randomArray, "localLookup: randomArray");
  xfree(buf, "localLookup: buf");

  return YES;
}

/**
 * Maximum length of the name of an indexed file (with path).
 **/ 
#define MAX_LINE_SIZE 1024

/* scan the list of on-demand shared files to
   initialize LOOUP_fileNames_ */
static int scanDatabaseList() {
  char * fil;
  FILE * handle;
  char * result;
  char * line;
  int pos;
  unsigned long long totalSize;

  if (indexed_files != NULL) {
    int i;
    for (i=0;i<indexed_files_size;i++)
      freeNonNull(indexed_files[i],
                  "freeing indexed file list");
    freeNonNull(indexed_files,
                "freeing old indexed_files");
  }
  indexed_files = NULL;
  indexed_files_count = 0;
  indexed_files_size = 0;
  fil = getSharedFileList();
  handle = fopen(fil, "a+");
  if (handle == NULL) {
    return SYSERR;
  }
  fseek(handle, 0, SEEK_SET);
  line = xmalloc(MAX_LINE_SIZE,
                 "space for filename in scanf");
  result = line;
  pos = 0;
  while (1) {
    result = fgets(line, MAX_LINE_SIZE - 1, handle);
    if (result == NULL)
      break;
    if (strlen(result) > 1)
      indexed_files_count++;
    pos++;
  }
  if (indexed_files_count == 0) {
    fclose(handle);
    xfree(line,
          "freeing space for scanf line");
    return OK;
  }
  fseek(handle, 0, SEEK_SET);
  indexed_files_size = pos;
  indexed_files = xmalloc(indexed_files_size * sizeof(char *),
                          "scanDatabaseList: filenameList");
  pos = 0;
  totalSize = 0;
  result = line;
  while (result != NULL) {
    result = fgets(line, MAX_LINE_SIZE - 1, handle);
    if ( result == NULL )
      break;
    if ( strlen(result) > 1 ) {
      line[strlen(line)-1] = '\0'; /* eat line feed */
      indexed_files[pos++] = strdup(line);
    } else
      indexed_files[pos++] = NULL;
  }
  xfree(line,
        "freeing space for scanf line");
  fclose(handle);
  return OK;
}

/**
 * Add a name to the list of filenames.
 * @returns the number of filenames in the index
 **/
int appendFilename(char * filename) {
  char * fil;
  FILE * handle;
  int result;
  char * line;
  char * scanf;
  unsigned short pos;

  if (filename == NULL) 
    errexit("appendFilename called with filename == NULL\n");
  fil = getSharedFileList();
  handle = fopen(fil, "r+");
  if (handle == NULL) 
    errexit("LOOKUP: List %s of directly shared filenames not available!\n",
	    fil);
  filename = expandFileName(filename);   
  if (filename == NULL)
    errexit("appendFilename - filename expand failed (returned NULL)\n");
  if (strlen(filename) >= MAX_LINE_SIZE) {
    fclose(handle);
    return SYSERR;
  }
  result = 0;
  pos = 0;
  line = xmalloc(MAX_LINE_SIZE,
		 "space for filename in scanf");
  while (1) {
    scanf = fgets(line, MAX_LINE_SIZE-1, handle);
    pos++;
    if (scanf == NULL)
      break;
    /* line always contains the line-feed, do not include that when
       diffing */
    if ( (strncmp(line, filename, strlen(filename)) == 0) &&
	 (strlen(line) == strlen(filename) + 1) ) {
      result = pos;
      break;
    }
  }
  xfree(line, 
	"freeing space for scanf line");
  if (result != 0) {
    fclose(handle);  
    xfree(filename, 
	  "appendFilename: expanded filename");  
    return result; /* already there! */
  }
  /* not there, append */
  fprintf(handle,
	  "%s\n", 
	  filename);
  fclose(handle);  
  xfree(filename, 
	"appendFilename: expanded filename");  
  scanDatabaseList();
  return pos; /* return index */
}

/**
 * Is space available on the drive? Returns the number of
 * blocks (1k) available. If there would be no space except
 * if less important content is discarded, this method 
 * *deletes* 1k of the less important content and returns 
 * 1. If there is only more important content, the
 * method returns 0.
 * @param importance how important is the content?
 **/
int spaceAvailable(int importance) {  
  if (LOOKUP_filesStored_ > getDiskQuota() * 1024) {
    if (importance+LOOKUP_age <= LOOKUP_leastImportance_)
      return 0; /* no */
    /* free lip entry */
    deleteEntry_(rootLevel, 
		 &LOOKUP_lip_);
    findLeastImportant(rootLevel);
    return 1;
  } else {
    LOOKUP_filesStored_++; /* it is safe to assume that there will be one more file very soon... */
    return getDiskQuota() * 1024 - LOOKUP_filesStored_; /* this many blocks are free */
  }     
}

/**
 * Cron-job that makes sure the lookup-module is fine
 * by synchronizing the in-memory database with the drive.
 **/
static void cronSyncLookup(CollisionLevel * level) {
  DEBUG("CRON: enter cronSyncLookup\n");
#if HAVE_FDATASYNC
  assert(fdatasync(level->handle) == 0,
	 "fdatasync failed on database!");
#else
  assert(fsync(level->handle) == 0,
	 "fdatasync failed on database!");
#endif
  DEBUG("CRON: exit cronSyncLookup\n");
}

/**
 * Cron-job that decreases the importance-level of all
 * files by 1. Runs 'not very often'.
 **/
static void cronReduceImportance(void * unused) {
 FileName fileName;
  int handle;
  
  DEBUG("CRON: enter cronReduceImportance\n");
  fileName= getAgeFile();
  handle = open(fileName,
		O_CREAT|O_RDWR,
		S_IRUSR|S_IWUSR);
  LOOKUP_age++;
  write(handle, 
       &LOOKUP_age,
       sizeof(int));
  close(handle);
  DEBUG("CRON: exit cronReduceImportance\n");
}

/**
 * Delete an entry and the corresponding file on the
 * drive.
 * @param entry the double-hash of the item to remove (=filename)
 **/
static void deleteEntry_(CollisionLevel * level,
			 HashCode160 * entry) {
  int noll = 0;
  HashCode160 tripleHash;
  int i;
  int j;
  size_t off;
  ContentIndex cmpEntry;

  if (1) {
    WARNING("deleteEntry_ not implemented!\n");
    return;
  }
  /* the implementation below is broken since it 
     does not check if entries from the collision-
     hashtables must be moved up into the free
     slot! */
  /* remove entry from database */
  hash(&entry,
       sizeof(HashCode160),
       &tripleHash);
  MUTEX_LOCK(&level->lock);
  off = getOffset(level, &tripleHash);
  lseek(level->handle, 
	off, 
	SEEK_SET);
  for (j=0;j<LOOKUP_CORUN;j++) {
    read(level->handle, 
	 &cmpEntry,
	 sizeof(ContentIndex));
    if (equalsHashCode160(entry,
			  &cmpEntry.doubleHash)) {      
      lseek(level->handle, 
	    off, 
	    SEEK_SET);
      /* fill database with zeros */
      for (i=0;i<sizeof(ContentIndex)/sizeof(int);i++)
	write(level->handle, &noll, sizeof(int));
      break; /* remove successful */
    }
    off += sizeof(ContentIndex);
  }
  MUTEX_UNLOCK(&level->lock);

  /* remove the file */
  unlinkFromDB(entry);
}

/**
 * Find the last important entry and
 * set the LOOKUP_lip_ and LOOKUP_leastImportance_
 * variables accordingly.
 **/
static void findLeastImportant(CollisionLevel * level) {
  ContentIndex entry;
  int j;
  int i;
  int k;
  int importances[LOOKUP_LIPS_COUNT];
 
  if (LOOKUP_filesStored_ < LOOKUP_LIPS_COUNT*LOOKUP_FACTOR)
    return;
  MUTEX_LOCK(&level->lock);
  while (LOOKUP_lips_count_ == 0) {
    LOOPPRINT("findLeastImportant");
    for (k=0;k<LOOKUP_LIPS_COUNT;k++)
      importances[k] = 0x7FFFFFFF;
    LOOKUP_lips_db_index_++;
    LOOKUP_lips_db_index_ = LOOKUP_lips_db_index_ % LOOKUP_LIPSPART_COUNT;
    lseek(level->handle, 
	  LOOKUP_lips_db_index_*(CACHE_HASH_TABLE_SIZE/LOOKUP_LIPSPART_COUNT)*sizeof(ContentIndex), 
	  SEEK_SET);
    for (i=LOOKUP_lips_db_index_*(CACHE_HASH_TABLE_SIZE/LOOKUP_LIPSPART_COUNT);
	 i<(LOOKUP_lips_db_index_+1)*(CACHE_HASH_TABLE_SIZE/LOOKUP_LIPSPART_COUNT);
	 i++) {
      j = read(level->handle, &entry, sizeof(ContentIndex));
      if (j == sizeof(ContentIndex)) {
	if (ntohl(entry.fileNameIndex) != -1) {
	  continue; /* on demand encoding or free */
	}
	for (k=0;k<LOOKUP_LIPS_COUNT;k++)
	  if (ntohl(entry.importance) < importances[k]) {
	    importances[k] = ntohl(entry.importance);
	    LOOKUP_lips_[k] = i;
	    break;
	  }
      }
    }
    for (k=0;k<LOOKUP_LIPS_COUNT;k++)
      if (importances[k] < 0x7FFFFFFF)
	LOOKUP_lips_count_++; /* found entry with low importance */
  }
  k = LOOKUP_lips_[--LOOKUP_lips_count_];
  lseek(level->handle, k*sizeof(ContentIndex), SEEK_SET);
  j = read(level->handle, &entry, sizeof(ContentIndex));
  if (j == sizeof(ContentIndex)) {
    LOOKUP_leastImportance_ = ntohl(entry.importance);
    memcpy(&LOOKUP_lip_,
	   &entry.doubleHash,
	   sizeof(HashCode160));      
  } /* else: warning: DB inconsistent? */
  MUTEX_UNLOCK(&level->lock);
}

/**
 * Lookup the file on the local machine. If found,
 * send it back.
 * @param query the triple-hash of the content
 * @param qp the policy (for us: priority)
 * @param callback procedure to call on the results,
 *        must take three arguments: NULL, a pointer
 *        to the CONTENT_Block, and the query.
 * @returns YES if found, NO if not.
 **/
int tripleHashLookup(HashCode160 * query,
		     QUERY_POLICY qp,
		     int (* callback)(HostIdentity*, CONTENT_Block*,
				       HashCode160*)) {
  ContentIndex ci;
  CollisionLevel * level;

  level = rootLevel;
  if (NO == findEntry_(&level,
		      query,
		      &ci)) {
    return NO;
  }
  ci.importance = htonl(ntohl(ci.importance) +
			(qp & QUERY_PRIORITY_BITMASK)); /* increase importance */
  writeEntry(level,
	     &ci,
	     NO); /* write to disk! */
  resolveEntry(&ci, callback); /* send reply */
  return YES; /* report success */
}

/**
 * Given an entry in the cache, obtain the matching
 * message(s). Send them to the given host.
 **/ 
static void resolveEntry(ContentIndex * entry,
			 int (* method)(HostIdentity*,CONTENT_Block*,
					HashCode160*)) {
  CONTENT_Block result;
  CONTENT_Block crypt;
  FileName fn;
  int fileHandle;
  ssize_t blen;
  HashCode160 hc;

  /* data/content mechanism */
  if (ntohl(entry->fileNameIndex) == -1) {
    localLookup(&entry->doubleHash,
		method);
    return;
  }
  if (ntohl(entry->fileNameIndex) > indexed_files_count) {
    FAILURE("Database inconsistent! (index points to invalid offset %d (>= %d))\n",
	    ntohl(entry->fileNameIndex), indexed_files_count);
    return;
  }
  /* on-demand encoding mechanism */
  fn = indexed_files[ntohl(entry->fileNameIndex)-1];
  fileHandle = open(fn, O_EXCL, S_IRUSR);
  if (fileHandle == -1) {
    FAILURE("Could not open file %s.\n",
	    fn);    
    return;
  }
  lseek(fileHandle, ntohl(entry->fileOffset), SEEK_SET);
  blen = read(fileHandle, 
	      &result,
	      sizeof(CONTENT_Block));
  close(fileHandle);
  hash(&result, 
       blen, 
       &hc);
  if (OK == encryptContent(&result,
			   &hc,
			   &crypt)) {
    /*    method = callback; */
    method(NULL, 
	   &crypt, 
	   &entry->doubleHash);  
  }
}

/**
 * Lookup the given Hash in the Hashtable-cache structure
 * (load blocks as needed).
 * @returns NO if not found, YES if found
 **/
int findEntry(HashCode160 * tripleHash,
	      ContentIndex * result) {
  CollisionLevel * root;

  root = rootLevel;
  return findEntry_(&root,
		    tripleHash,
		    result);
}

/**
 * Lookup the given Hash in the Hashtable-cache structure
 * (load blocks as needed).
 * @returns NO if not found, YES if found
 **/
static int findEntry_(CollisionLevel ** level,
		      HashCode160 * query,
		      ContentIndex * entry) {
  HashCode160 tripleHash;
  int j;
  int emptySlot = NO;
  int size;

  MUTEX_LOCK(&(*level)->lock);
  lseek((*level)->handle, 
	getOffset(*level, query), SEEK_SET);
  entry->fileNameIndex = 0;
  for (j=0;j<LOOKUP_CORUN;j++) {
    size = read((*level)->handle, 
		entry, 
		sizeof(ContentIndex));
    if ( (size == 0) || 
	 (entry->fileNameIndex == 0) ) 
      emptySlot = YES;
    hash(&entry->doubleHash,
	 sizeof(HashCode160),
	 &tripleHash);
    if (equalsHashCode160(query, 
			  &tripleHash)) {
      MUTEX_UNLOCK(&(*level)->lock);
      return YES;
    }
  }
  MUTEX_UNLOCK(&(*level)->lock);
  if (emptySlot == YES) {
    DEBUG("INFOMERCIAL: nothing found locally!\n");
    return NO; /* no collision, nothing there */
  }
  if ((*level)->next != NULL) {
    level = &((*level)->next);
    return findEntry_(level,
		      query,
		      entry);
  } else
    return NO;
}

/**
 * Call a method for each entry in the lookup table.
 * VERY expensive (will take quite a while to run),
 * so don't do this unless under exceptional circumstances.
 * @param method the method to call, abort early if 
 *        method returns SYSERR
 * @return the number of entries in the database
 **/
int forEachLookupEntry(int (*method)(ContentIndex * entry,
				     CollisionLevel * level,
				     size_t pos)) {
  CollisionLevel * current;
  size_t pos;
  ContentIndex entry;
  unsigned int count;
  
  count = 0;
  current = rootLevel;
  while (current != NULL) {
    pos = 0;
    while (pos < current->mask) {
      MUTEX_LOCK(&current->lock);  
      lseek(current->handle, 
	    pos*sizeof(ContentIndex), 
	    SEEK_SET);
      if (sizeof(ContentIndex) != read(current->handle,
				       &entry,
				       sizeof(ContentIndex)))
	break; /* end of file */
      MUTEX_UNLOCK(&current->lock);  
      if ( (entry.fileNameIndex != 0) ||
	   (entry.fileOffset != 0) ) {
	count++;
	if (SYSERR == method(&entry, 
			     current,
			     pos)) {
	  return count;
	}
      }
      pos++;
    }
    current = current->next;
  }
  return count;
}

/**
 * Write the given entry to the database. 
 * @return OK on success, SYSERR on error
 **/
int addEntry(ContentIndex * entry) {
  return addEntry_(rootLevel, entry);
}

/**
 * Write the given entry to the database. 
 * @return OK on success, SYSERR on error
 **/
static int addEntry_(CollisionLevel * level,
		     ContentIndex * entry) {
  int size;
  HashCode160 tripleHash;
  ContentIndex onDisk;
  int hasFreeBlock = NO;
  int j;

  hash(&entry->doubleHash,
       sizeof(HashCode160),
       &tripleHash);
  entry->importance = htonl(ntohl(entry->importance)+LOOKUP_age);
  MUTEX_LOCK(&level->lock);
  lseek(level->handle, 
	getOffset(level, &tripleHash), 
	SEEK_SET);
  for (j=0;j<LOOKUP_CORUN;j++) {    
    size = read(level->handle, 
		&onDisk, 
		sizeof(ContentIndex));
    if ( (size == 0) || 
	 (onDisk.fileNameIndex == 0) ) 
      hasFreeBlock = YES;
    if (equalsHashCode160(&onDisk.doubleHash, 
			  &entry->doubleHash)) {
      MUTEX_UNLOCK(&level->lock);
      writeEntry(level, 
		 entry, 
		 NO); /* update */
      return OK;
    }
  }    
  MUTEX_UNLOCK(&level->lock);
  if ( hasFreeBlock == YES ) {
    writeEntry(level, entry, YES);
    return OK; /* no collision, nothing there, just write */
  }
  if (level->next != NULL)
    return addEntry_(level->next,
		     entry);
  else
    return SYSERR; /* all collision levels full! */
}

/**
 * Write the given entry to the database.
 * @param isNew is NO for updates, YES for new entries
 **/
static void writeEntry(CollisionLevel * level,
		       ContentIndex * entry,
		       int isNew) {
  HashCode160 tripleHash;
  int j;
  int size;
  size_t off;
  ContentIndex onDisk;

  /*  printf("in writeEntry %d\n", isNew);*/
  hash(&entry->doubleHash,
       sizeof(HashCode160),
       &tripleHash);
  
  off = getOffset(level,
		  &tripleHash);
  MUTEX_LOCK(&level->lock);
  lseek(level->handle, 
	off,
	SEEK_SET);
  for (j=0;j<LOOKUP_CORUN;j++) {
    size = read(level->handle, 
		&onDisk, 
		sizeof(ContentIndex));
    if ( (size == 0) ||
	 ( (onDisk.fileNameIndex == 0) &&
	   (isNew == YES) ) ) {
      lseek(level->handle, 
	    off,
	    SEEK_SET);
      write(level->handle, 
	    entry, 
	    sizeof(ContentIndex));
      break;
    }
    if (equalsHashCode160(&onDisk.doubleHash, 
			  &entry->doubleHash)) {
      lseek(level->handle, 
	    off,
	    SEEK_SET);
      write(level->handle, 
	    entry, 
	    sizeof(ContentIndex));
      break;
    }
    off+= sizeof(ContentIndex);
  }
  if (j == LOOKUP_CORUN)
    FAILURE("ASSERT FAILED: invalid call to writeEntry?");
  MUTEX_UNLOCK(&level->lock);
}


/* end of lookup.c */
