/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/** 
 * The RequestManager keeps track and re-issues
 * requests for Content-Nodes.
 * @file textui/requestmanager.c
 * @author Christian Grothoff
 **/ 

#include "config.h"
#include "debugging.h"
#include "util/logging.h"
#include "util/checksum.h"
#include "textui/requestmanager.h"
#include "configuration.h"
#include "knownhosts.h"
#include <math.h>
#include <pthread.h>


/* how often do we try a query  (in a row, it will be
   re-scheduled later!) */
#define MAX_TRIES 5

static void performRequest(RequestEntry * request,
			   RequestThreadArgs * args) {
  HexName hn;
  int liveQueries;
  int i;
  time_t now;
  unsigned int ttl;
 
  time(&now);
  if (request->lasttime + ntohl(request->message.ttl) > now) {
    hash2hex(&request->message.query,
	     &hn);
    DEBUG("Delaying request for %s with ttl %d (remaining: %d)\n",
	  (char*)&hn, 
	  ntohl(request->message.ttl),
	  request->lasttime + ntohl(request->message.ttl) - now);
    return;
  }
  hash2hex(&request->message.query,
	   &hn);
  liveQueries = 0;
  for (i=0;i<MAX_REQUESTS;i++)
    if ( args->REQUESTMANAGER_requestedList_[i] != NULL)
      liveQueries++;
  DEBUG("asking for %s with ttl %d and priority %d (retry: %d) Pending: %d/%d (%d)\n",
	(char*)&hn, 
	ntohl(request->message.ttl),
	ntohl(request->message.priority),
	request->tries,
	args->REQUESTMANAGER_sendRequestCount_,
	args->REQUESTMANAGER_requestListSize_,
	liveQueries);

  ttl = ntohl(request->message.ttl);
  ttl = ttl*2 + randomi(4*TTL_DECREMENT);
  i = forEachHost(NULL, 0, NULL); /* estimate network size! */
  if (i<3) i=3; /* ensure log is >= 1 */
  /* bound the TTL by 'two TTL_DECREMENT' per log(#hosts)' */
  if (ttl > 4 * TTL_DECREMENT * (int) log(i))
    ttl = 4 * TTL_DECREMENT * ((int)log(i)) + randomi(2*TTL_DECREMENT);
  request->message.ttl = htonl(ttl);

  writeToSocket(args->REQUESTMANAGER_socket_,
		(TCP_SOCKET_BUFFER*)&request->message);	     
  request->message.priority 
    = htonl(ntohl(request->message.priority)*2+randomi(2*TTL_DECREMENT));
  request->lasttime = now;

  /*#define MACRO(a,b) a = htonl(ntohl(a)b)
    MACRO(request->message.priority,+1);*/
  
  request->tries++;
  if ( (request->tries > MAX_TRIES) &&
       (args->REQUESTMANAGER_sendRequestCount_ > 0) ){
    RequestEntry tmp;
    RequestEntry * target;
    request->tries = 0;
    memcpy(&tmp,
	   request,
	   sizeof(RequestEntry));
    target = args->REQUESTMANAGER_requestList_
      [ randomi(args->REQUESTMANAGER_sendRequestCount_)];
    memcpy(request,
	   target,
	   sizeof(RequestEntry));
    memcpy(target,
	   &tmp,
	   sizeof(RequestEntry));
  }
}

static void cronRequestManager(void * data) {
  int i;
  int active;
  RequestEntry * send_i;
  RequestThreadArgs * args;

  args = data;

  DEBUG("CRON: enter cronRequestManager\n");
  active = 0;
  MUTEX_LOCK(&args->REQUESTMANAGER_Lock_);
  for (i=0;i<MAX_REQUESTS;i++) {
    send_i = args->REQUESTMANAGER_requestedList_[i];
    if (send_i != NULL) {
      performRequest(send_i,
		     args);    
      active++;
    } else
      DEBUG("request manager has no request pending at slot %d (%d)\n",
	    i,args->REQUESTMANAGER_sendRequestCount_);
  }
  if ( (isOpenConnection(args->REQUESTMANAGER_socket_)) &&
       (active == 0) ) {
    DEBUG("request manager has no pending requests, closing socket %d.\n",
	  args->REQUESTMANAGER_socket_->socket);
    closeSocketTemporarily(args->REQUESTMANAGER_socket_);
  }
  MUTEX_UNLOCK(&args->REQUESTMANAGER_Lock_);
  DEBUG("CRON: exit cronRequestManager\n");
}

void initRequestManager(char * hostname,
			RequestThreadArgs ** args) {
  int i;

  *args=xmalloc(sizeof(RequestThreadArgs),
  	"initRequestManager: args");

  (*args)->REQUESTMANAGER_socket_ = xmalloc(sizeof(GNUNET_TCP_SOCKET),
			    "initRequestManager: REQUESTMANAGER_socket_");

  initGNUnetClientSocket(getGNUnetPort(),
			 hostname,
			 (*args)->REQUESTMANAGER_socket_);
  (*args)->REQUESTMANAGER_initialTTL_ = 1+2*TTL_DECREMENT;
  (*args)->REQUESTMANAGER_duplicationEstimate_ = 0;
  create_recursive_mutex(&(*args)->REQUESTMANAGER_Lock_);
  create_recursive_mutex(&(*args)->REQUESTMANAGER_ReadLock_);
  (*args)->REQUESTMANAGER_sendRequestCount_ = 0;
  (*args)->REQUESTMANAGER_requestListSize_ = MAX_REQUESTS;
  (*args)->REQUESTMANAGER_requestList_ 
    = (RequestEntry **) xmalloc
    ((*args)->REQUESTMANAGER_requestListSize_*sizeof(RequestEntry*),
     "initRequestManager: requestList");
  for(i=0;i<MAX_REQUESTS;i++)
    (*args)->REQUESTMANAGER_requestedList_[i]=NULL;
  
  addCronJob(&cronRequestManager, 
             1000*1, 1000*1,
	     *args);
}

/**
 * This implementation is only half-baked.
 * We may also want to free a couple of the request
 * datastructures and do other things.
 **/
void doneRequestManager(RequestThreadArgs * args) {
  delCronJob(&cronRequestManager, 
             1000*1,
	     args);
  destroySocket(args->REQUESTMANAGER_socket_);

  xfree(args->REQUESTMANAGER_socket_,
        "doneRequestManager: REQUESTMANAGER_socket_");
  
  destroy_mutex(&args->REQUESTMANAGER_Lock_);
  destroy_mutex(&args->REQUESTMANAGER_ReadLock_);
  xfree(args->REQUESTMANAGER_requestList_,
	"doneRequestManager: requestList");
  xfree(args,
  	"doneRequestManager: args");
}

/**
 * This method receives data corresponding to the indicated filename (hashcode)
 * Find the Listener that scheduled this request and
 * drop it from the list of pending requests.
 * @param hash hashcode representing filename
 * @param data data corresponding to this node
 **/ 
void requestManagerReceive(HashCode160 * hash, 
			   CONTENT_Block * data,
			   RequestThreadArgs * args) {
  int i;
  RequestEntry * send_i;
  time_t currentTime;
  HexName hn;
  int discarded;
  
  hash2hex(hash,
	   &hn);
  time(&currentTime);
  discarded = 0;
  MUTEX_LOCK(&args->REQUESTMANAGER_Lock_);
  for (i=0;i<MAX_REQUESTS;i++) {
    send_i = args->REQUESTMANAGER_requestedList_[i];
    if (send_i == NULL)
      continue;
    if (equalsHashCode160(hash, &send_i->message.query)) {
      discarded++;
      args->REQUESTMANAGER_initialTTL_ 
	= ((((int) ((currentTime-send_i->lasttime))) +
	   (args->REQUESTMANAGER_initialTTL_*3))) / 4;
      if (args->REQUESTMANAGER_initialTTL_ < TTL_DECREMENT)
	args->REQUESTMANAGER_initialTTL_ = TTL_DECREMENT; /* lower bound! */
      if (args->REQUESTMANAGER_initialTTL_ > 20 * TTL_DECREMENT)
	args->REQUESTMANAGER_initialTTL_ = 20 * TTL_DECREMENT;
      if (SYSERR == (((int (*)(Node*,HashCode160 *,CONTENT_Block*,RequestThreadArgs *))
		      send_i->receiver))(send_i->metadata,hash,data,args)) {
	MUTEX_UNLOCK(&args->REQUESTMANAGER_Lock_);
	return;
      }	
      xfree(send_i,
	    "requestManagerReceive: requestEntry");
      send_i = NULL;
      
      if (args->REQUESTMANAGER_sendRequestCount_ > 0) {
	args->REQUESTMANAGER_sendRequestCount_--;
	send_i = args->REQUESTMANAGER_requestList_[args->REQUESTMANAGER_sendRequestCount_];
	send_i->lasttime = 0;
	send_i->message.priority = htonl(5+randomi(20));
	send_i->message.header.size = htons(sizeof(TCP_Query_Request));
	send_i->message.header.tcpType = htons(TCP_REQUEST_QUERY);
	send_i->message.ttl = htonl((randomi(args->REQUESTMANAGER_initialTTL_)+TTL_DECREMENT));
 	performRequest(send_i,
		       args); /* request now - no wait! */
      }
      args->REQUESTMANAGER_requestedList_[i] = send_i;
    }
  }
  MUTEX_UNLOCK(&args->REQUESTMANAGER_Lock_);  
  if (discarded == 0) {
    args->REQUESTMANAGER_duplicationEstimate_ ++;
    DEBUG("RequestManager: received useless data %d (%d)!\n",
	  args->REQUESTMANAGER_duplicationEstimate_ ,
	  args->REQUESTMANAGER_initialTTL_);
  }
}

/**
 * Start download thread that will exit once the
 * download for the given root-node is complete.
 *
 * @param root the file to download
 * @param data 3rd argument to the method specified in root->downloadModel
 **/
void requestManagerReceiveResults(RNode * root,
				  void * data,
				  RequestThreadArgs * args) {
  TCP_SOCKET_BUFFER buffer;
  TCP_Content_Reply * reply;
  int timeout;
  HexName hex;
  time_t startTime;
  time_t now;
  float bps = 0;
  unsigned int togo = 0;

  reply = (TCP_Content_Reply *) &buffer;
  timeout = 1;
  time(&startTime);
  while (root->bytes_downloaded < root->fileLength) {
    if (SYSERR == readFromSocket(args->REQUESTMANAGER_socket_,
				 &buffer)) {
      sleep(timeout);
      timeout+=2;
      continue;
    }
    switch (ntohs(reply->header.tcpType)) {
    case TCP_REPLY_CONTENT:
     if (ntohs(reply->header.size) != sizeof(TCP_Content_Reply)) {
       fprintf(stderr,"Received bad reply: %d\n",
	       ntohs(reply->header.tcpType));
   	closeSocketTemporarily(args->REQUESTMANAGER_socket_);
	break;
      }
      timeout = 1; 
      hash2hex(&reply->hash, &hex);
      if (root->bytes_downloaded == 0)
	time(&startTime);
      time(&now);
      if (now > startTime) {
	bps = (root->bytes_downloaded / (now-startTime));
	if (bps > 0.001) {
	  togo = (unsigned int) ( (root->fileLength - root->bytes_downloaded) / bps);
	} else
	  togo = (unsigned int) -1;
      } else {
	bps = 0;
	togo = 0;
      }
      INFO("%8d of %8d (last: %s) at %8u bps (average), %4us to go\n",
	   root->bytes_downloaded, root->fileLength,
	   (char*)&hex,
	   (unsigned int)bps,
	   togo);
      requestManagerReceive(&reply->hash,
			    &reply->content,
			    args);    
      if (root->downloadModel != NULL) {
    
	MUTEX_LOCK(&args->REQUESTMANAGER_ReadLock_);
	((void*(*)(size_t, size_t, void*)) root->downloadModel)(root->bytes_downloaded,
								root->fileLength,
								data);
	MUTEX_UNLOCK(&args->REQUESTMANAGER_ReadLock_);
      }
     break;
    default:
      WARNING("TCP: Received unknown type: %d\n",
	      ntohs(reply->header.tcpType));
      /* ignore */
      break;
    }
  }
  /*  destroySocket(&REQUESTMANAGER_socket_);*/
}

/**
 * Queue a request for execution.
 **/
void requestManagerRequest(Node * node,
			   Listener l,
			   RequestThreadArgs * args) {
  int i;
  time_t currentTime;
  RequestEntry ** tmp;
  RequestEntry * send_i;
  HashCode160 * query;

  query = &node->request;
  time(&currentTime);
  MUTEX_LOCK(&args->REQUESTMANAGER_Lock_);
  for (i=0;i<MAX_REQUESTS;i++) { 
    send_i = args->REQUESTMANAGER_requestedList_[i];
    if (send_i != NULL) 
      continue;
    send_i = xmalloc(sizeof(RequestEntry),
		     "requestManagerRequest (requestEntry)");
    args->REQUESTMANAGER_requestedList_[i] = send_i;
    send_i->receiver = l;
    memcpy(&send_i->message.query,
	   query,
	   sizeof(HashCode160));
    send_i->lasttime = 0;
    send_i->metadata = node;
    send_i->message.header.size = htons(sizeof(TCP_Query_Request));
    send_i->message.header.tcpType = htons(TCP_REQUEST_QUERY);
    send_i->message.ttl = htonl(randomi(args->REQUESTMANAGER_initialTTL_)+TTL_DECREMENT);
    send_i->message.priority = htonl(1);
    send_i->tries = 0;
    performRequest(send_i,
		   args);
    MUTEX_UNLOCK(&args->REQUESTMANAGER_Lock_);
    return;
  }	    
  /* do we have to grow the queue? */
  if (args->REQUESTMANAGER_requestListSize_ == 
      args->REQUESTMANAGER_sendRequestCount_) {
    tmp = args->REQUESTMANAGER_requestList_;
    args->REQUESTMANAGER_requestList_ 
      = (RequestEntry **) xmalloc
      (2*args->REQUESTMANAGER_requestListSize_*sizeof(RequestEntry*),
       "requestManagerRequest. requestList");
    memcpy(args->REQUESTMANAGER_requestList_, tmp,
	   args->REQUESTMANAGER_requestListSize_*sizeof(RequestEntry*));
    args->REQUESTMANAGER_requestListSize_ *= 2;
  }
  /* append to the queue */
  args->REQUESTMANAGER_requestList_
    [ args->REQUESTMANAGER_sendRequestCount_]
    = xmalloc(sizeof(RequestEntry),
	      "requestManagerRequest: requestEntry (2)");
  send_i = args->REQUESTMANAGER_requestList_
    [ args->REQUESTMANAGER_sendRequestCount_];
  args->REQUESTMANAGER_sendRequestCount_++;
  send_i->receiver = l;
  memcpy(&send_i->message.query,
	 query,
	 sizeof(HashCode160)); 
  send_i->lasttime = 0;
  send_i->metadata = node;
  send_i->message.header.size = htons(sizeof(TCP_Query_Request));
  send_i->message.header.tcpType = htons(TCP_REQUEST_QUERY);
  send_i->message.ttl = htonl(randomi(args->REQUESTMANAGER_initialTTL_)+TTL_DECREMENT);
  send_i->message.priority = htonl(1);
  send_i->tries = 0;
  MUTEX_UNLOCK(&args->REQUESTMANAGER_Lock_);
}
  
/**
 * Abort a request
 * @param node what was the query
 **/
void requestManagerAbort(Node * node,
			 RequestThreadArgs * args) {
  int i;
  MUTEX_LOCK(&args->REQUESTMANAGER_Lock_);
  for (i=0;i<MAX_REQUESTS;i++) {
    if (args->REQUESTMANAGER_requestedList_[i]->metadata == node) {
      xfree(args->REQUESTMANAGER_requestedList_[i],
	    "requestManagerAbort: requestedList");
      
      if (args->REQUESTMANAGER_sendRequestCount_ > 0) 
	args->REQUESTMANAGER_requestedList_[i] 
	  = args->REQUESTMANAGER_requestList_[--(args->REQUESTMANAGER_sendRequestCount_)];
    }
  }
  for (i=0;i<args->REQUESTMANAGER_requestListSize_;i++)
    if (args->REQUESTMANAGER_requestList_[i]->metadata == node) {
      xfree(args->REQUESTMANAGER_requestList_[i],
	    "requestManagerAbort: requestList");
      args->REQUESTMANAGER_requestList_[i] 
	= args->REQUESTMANAGER_requestList_[--(args->REQUESTMANAGER_sendRequestCount_)];
    }
  MUTEX_UNLOCK(&args->REQUESTMANAGER_Lock_);  
}



/* end of requestmanager.c */
